"""Contains all the defined mloq Commands."""
from mloq.commands.ci import CiCMD
from mloq.commands.docker import DockerCMD
from mloq.commands.docs import DocsCMD
from mloq.commands.globals import GlobalsCMD
from mloq.commands.license import LicenseCMD
from mloq.commands.lint import LintCMD
from mloq.commands.package import PackageCMD
from mloq.commands.project import ProjectCMD
from mloq.commands.requirements import RequirementsCMD
from mloq.commands.setup import SetupCMD
